/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <string.h>                         /* strstr() */
#include <stdlib.h>                         /* free() exit() */
#include <assert.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_env_variable.h"

#include "initng_conflict.h"

s_entry CONFLICT = { "conflict", STRINGS, NULL,
    "If service put here is starting or running, bail out."
};

static int check_conflict(active_db_h * service,
                                     a_state_h * status)
{
    const char *conflict_entry = NULL;
    
    assert(service);
    assert(service->name);

    /* make sure the conflict entry is set */
    while((conflict_entry=initng_active_db_get_next_string(&CONFLICT, service, conflict_entry)))
    {
	active_db_h *s = NULL;
	
	/*D_("Making sure that %s is not running.\n", conflict_entry);*/
	char *fixed = fix_variables(conflict_entry, service);
	
	s=initng_active_db_find_by_name(fixed);

	free(fixed);
	/* this is actually good */	
	if(!s)
	{
	    /*D_("Conflict not found!\n");*/
	    continue;
	}
	
	if(IS_UP(s) || IS_STARTING(s))
	{
	    initng_common_mark_service(service, &FAIL_STARTING);
	    F_("Service \"%s\" is conflicting with service \"%s\"!\n", service->name, s->name);
	    return(FALSE);
	}
    }    
    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_service_data_types_add(&CONFLICT);
    initng_active_state_add_test(&LOADING, 10,
                                 &check_conflict);

    return (TRUE);
}

void module_unload(void)
{

    initng_active_state_del_test(&LOADING, &check_conflict);
    initng_service_data_types_del(&CONFLICT);
}
