
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <dirent.h>   /*#include <dirent.h> */  /* opendir() closedir() */
#include <fnmatch.h>

#include "../../src/initng_global.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_error.h"
#include "../../src/initng_handler.h"
#include "../../initng-paths.h"

#include "initng_i_parser.h"

#define ENDED(x) (!(x)[0] || (x)[0]=='\n')
char *g_filename = NULL;
char *g_pointer = NULL;         /* a pointer to the first char we can backstep to when printing line copys */


/* set this to have a really verbose walking output */
/* #define DL_(a,m) err_print_line2t(a, m, MSG, __FILE__, (const char*)__PRETTY_FUNCTION__, __LINE__)*/
#define DL_(a,m)

#define FL_(a,m) err_print_line2t(a, m, MSG_FAIL, __FILE__, (const char*)__PRETTY_FUNCTION__, __LINE__)

static void err_print_line2t(char *point, const char *message, e_mt err,
                             const char *file, const char *func,
                             int codeline);
static service_cache_h *parse_file(char *filename, const char *watch_for);
static int parse_service_line(char **to_parse, const char *watch_for,
                              service_cache_h * father, stype_h * type,
                              char *filename, service_cache_h ** match,
                              service_cache_h ** exact_match);
static int parse_opt(char **where, stype_h * type, service_cache_h * srv);

static int set_parser(s_entry * type, char **value, char *va,
                      service_cache_h * from_service);
static int string_parser(s_entry * type, char **value, char *va,
                         service_cache_h * from_service);
static int strings_parser(s_entry * type, char **value, char *va,
                          service_cache_h * from_service);
static int iint_parser(s_entry * type, char **value, char *va,
                       service_cache_h * from_service);

static char *dup_string_and_walk(char **value, int break_on_space);

static int is_valid(char *string);

static int is_valid(char *string)
{
    int i;

    for (i = 0; string[i] != 0; i++)
        if (string[i] <= 32)
            return (FALSE);

    return (TRUE);
}

static void err_print_line2t(char *point, const char *message, e_mt err,
                             const char *file, const char *func, int codeline)
{
    char *line;
    char *pstr;
    int len = 0;
    int i = 0;
    int pi = 0;
    char *start = point;

    /* make sure global filename is set */
    if (!g_pointer || !g_filename)
        return;

    /*while(start != g_pointer && start[-1] != g_pointer && start[-1] != '\n') */
    while (start != g_pointer && (start - 1)[0] != '\n')
        start--;

    while (start[len] && start[len] != '\n')
        len++;

    pi=point-start;

    /* copy line working on */
    line = i_strndup(start, len);
    pstr = i_calloc(pi+4, sizeof(char));
    if (!line || !pstr) return;

    /* put spaces */    
    memset(pstr, ' ', sizeof(char) * (pi) );
    
    /* Tabs destroy our posistion pointers, replaces with spaces */
    for(i=0;line[i];i++)
	if(line[i]=='\t')
	    line[i]=' ';


    /* then, put the marker. */
    pstr[pi] = '^';
    pstr[pi + 1] = '\0';


    initng_error_print(err, file, func, codeline,
                       "%s\nfile: %s\n\"%s\"\n %s\n\n", message,
                       g_filename, line, pstr);
    free(line);
    free(pstr);
    sleep(0.2);
}



/* Load a service from a service_to_find or process_path */
service_cache_h *initng_i_parser(const char *service_to_find)
{
    char filetoparse[200];
    service_cache_h *got_serv = NULL;
    char *path = NULL;
    const char *service = NULL;

    assert(service_to_find);

    /* Make sure the filename of 200 chars abow are more than enought */
    if (strlen(service_to_find) > 50)
    {
        F_("Service name to long, initng_i_parser can't look for this service!\n");
        return (NULL);
    }

    /*
     * now set up path and service.
     * if service_to_find is system/initial/udevd
     * path should be "system/initial"
     * and service "udevd"
     */

    /* set up a patch copy, has to be free() */
    path = st_get_path(service_to_find);

    /* this is a pointer to a entry in service_to_find */
    service = st_strip_path(service_to_find);

    D_("initng_i_parser(%s);\n", service_to_find);
    /* example  INITNG_ROOT=/etc/initng service_to_find="deamon/sshd INITNG_EXT=".i" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.i" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.i" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.i" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.i */
    /* TRY NO 6, try load   "/etc/initng/sshd.i" */
    /* TRY NO 7, try load   "/etc/initng/daemon.i */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/");
    strcat(filetoparse, service);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/default" INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 4, try "/etc/initng/daemon/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, "/default" INITNG_EXT);

    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, "/");
    strcat(filetoparse, path);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 6, try "/etc/initng/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 7, try "/etc/initng/daemon.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    D_("Was not able to parse: %s\n", service_to_find);
    free(path);
    return (NULL);
}

/* parse a file for services */
static service_cache_h *parse_file(char *filename, const char *watch_for)
{
    /* Pointer to content read, and pointer to where we are reading */
    char *file_content = NULL;
    char *file = NULL;


    /* if service is matching, it is set, if service is exeact_match this is set */
    service_cache_h *match = NULL;
    service_cache_h *exact_match = NULL;

    /* standard assertions */
    assert(filename);
    assert(watch_for);

    /* set the error message filename */
    g_filename = filename;

    D_("parse_file(%s);\n", filename);

    /* read config file */
    if (!open_read_close(filename, &file_content))
    {
        D_("Can't open config file %s.\n", filename);
        return (NULL);
    }
    file = file_content;
    g_pointer = file;


    /* main parse line by line loop */
    while (file[0] != '\0')
    {
        stype_h *stype = NULL;

        DL_(file, "parse_file while loop:");
        /* skip leading spaces */
        JUMP_NSPACES(file);
        if (file[0] == '\0')
	{
	    DL_(file, "End of file");
            break;
	}

        /* skip lines starting with '#' */
        if (file[0] == '#' || file[0] == '\n' || file[0] == ';')
        {
            REALLY_JUMP_TO_NEXT_ROW(file);
            continue;
        }
        /*
         * parsometer:
         *   service test {
         *   ^
         */

        /* get the stype */
        if (!(stype = initng_service_types_get(file)))
        {
            match = NULL;
            break;
        }

        JUMP_TO_NEXT_WORD(file);
        /*
         * parsometer:
         *   service test {
         *           ^
         */

        /* parse line, with options from that service type */
        if (parse_service_line(&file, watch_for, NULL, stype,
                               filename, &match, &exact_match) == TRUE)
        {
            /* if we found RIGHT what we wanted */
            if (exact_match)
            {
                match = exact_match;
                break;
            }
	    
	    /* ok, jump the ending '}' stack */
	    if(file[0] == '}')
		file++;
        }

    }                                       /* end while */

    if (file_content)
    {
        free(file_content);
        file_content = NULL;
    }

    /* add to service db */
    if (match && !initng_service_cache_add(match))
    {
	/* Problem adding, probably this service name exists allredy,
	 * DO NOT EVEN TRY TO UNLOAD EXISTANT! some services my already rely on it or 
	 * we can get here if we processing 'need' option from this service, unloading
	 * will make watch_for to point on not existing area of memory!
	 */
	service_cache_h *found;
	
	/* search for an with same name */
	found = initng_service_cache_find_by_name(match->name);
	
	/* if found, free created an return founded */
	if(found)
	{
    	    initng_service_cache_free(match);
	    match = found;
	} else if (!initng_service_cache_add(match))
	{
    	    F_("Could not add service to service db!\n");
    	    initng_service_cache_free(match);
    	    match = NULL;
	}
    }

#ifdef DEBUG
    if (!match)
	D_("%s - no match", watch_for);
#endif

    /* Return what we got, and hope for the best */
    return (match);
}


static int parse_service_line(char **to_parse, const char *watch_for,
                              service_cache_h * father, stype_h * type,
                              char *filename, service_cache_h ** match,
                              service_cache_h ** exact_match)
{
    char *name = NULL;
    char *father_name = NULL;
    service_cache_h *new_service = NULL;
#ifdef SUB_CLASS
    service_cache_h *duplicate = NULL;
    int len;
#endif

    assert(to_parse);
    assert(*to_parse);
    assert(watch_for);

    /* for debuging */
    DL_(*to_parse, "parse_service_line:");

    /* service test : class { */
    /*        |               */
    /* jump forward on spaces */
    JUMP_NSPACES(*to_parse);
    if (ENDED(*to_parse))
    {
        FL_(*to_parse, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* service test : class { */
    /*         |              */

    /* this fetches name, and increase *to_parse */
    if (!(name = st_dup_next_word(to_parse)))
    {
        FL_(*to_parse, "Did not get a name!");
        return (FALSE);
    }

    if (!is_valid(name))
    {
        FL_(*to_parse, "Name contains invalid characters.");
        return (FALSE);
    }

    /* service test : class { */
    /*             |          */

    D_("parse_service_line(s,%s): service name: \"%s\"\n", watch_for, name);

    /* jump forward on spaces */
    JUMP_NSPACES(*to_parse);
    if (ENDED(*to_parse))
    {
        FL_(*to_parse, "Line ended unexpectedly.");
        free(name);
        return (FALSE);;
    }

    /* service test : class { */
    /*              |         */
    if ((*to_parse)[0] == ':')
    {
        (*to_parse)++;
        /* service test : class { */
        /*               |        */

        /* jump forward on spaces */
        JUMP_NSPACES(*to_parse);
        if (ENDED(*to_parse))
        {
            FL_(*to_parse, "Line ended unexpectedly.");
            free(name);
            return (FALSE);
        }
        /* service test : class { */
        /*                |       */

        if (!(father_name = st_dup_next_word(to_parse)))
        {
            FL_(*to_parse, "Unable to fetch fathername.");
            free(name);
            return (FALSE);
        }

        if (!is_valid(father_name))
        {
            FL_(*to_parse,
                           "Father name contains invalid characters.");
            return (FALSE);
        }

#ifdef DEBUG
        if (father_name)
            D_("Father is: %s, set from \"server : father\"\n", father_name);
#endif

        /* service test : class { */
        /*                     |  */
        /* jump forward on spaces */
        JUMP_NSPACES(*to_parse);
        if (ENDED(*to_parse))
        {
            FL_(*to_parse, "Line ended unexpectedly.");
            free(name);
            free(father_name);
            return (FALSE);
        }
    }

    DL_(*to_parse, "Parse opt, shud stand direcltly over { char");
    /* service test : class { */
    /*                      | */
    if ((*to_parse)[0] != '{')
    {
        /* something is wrong */
        FL_(*to_parse, "error parsing, expected something else.");
        free(name);
        if (father_name)
            free(father_name);
        return (FALSE);
    }

    /* jump to first char after start tag and begin from there. */
    (*to_parse)++;
    JUMP_NSPACES(*to_parse);

    if (initng_service_cache_find_by_exact_name(name))
    {
	D_("service %s already parsed, skipping to the end of stack\n", name);
	int stack=1;
	int quoted = FALSE;
	int backslashed = FALSE;
	while ((*to_parse)[0])
	{
	    if (backslashed == FALSE)
	    {
		if ((*to_parse)[0] == '\\')
		    backslashed = TRUE;
		else if ((*to_parse)[0] == '"' || (*to_parse)[0] == '\'')
		{
		    if (quoted == FALSE)
			quoted = (*to_parse)[0];
		    else if (quoted == (*to_parse)[0])
			quoted = FALSE;
		} 
		else if (quoted == FALSE)
		{
		    if ((*to_parse)[0] == '{')
			stack++;
		    else if ((*to_parse)[0] == '}')
			stack--;
		    else if ((*to_parse)[0] == '#')
		    {
			REALLY_JUMP_TO_NEXT_ROW(*to_parse);
			continue;
		    }
		}
	    } else
		backslashed = FALSE;

	    if (stack == 0)
		break;

	    (*to_parse)++;
	}

	free(name);
	if (father_name)
	    free(father_name);
	
	return (TRUE);
    }

    DL_(*to_parse, "Parse opt, shud be first char in stack");
    

    /* service test : class { */
    /*                       | */

    new_service = initng_service_cache_new(name, type);

    /* from here, dont use name anymore, use new_service->name */
    free(name);
    name = NULL;

    /* FROM NOW, we should not free name or father_name, becouse its ADDED, and required */

    /* check so that it was allocated! */
    if (!new_service)
    {
        F_("Unable to allocate space for new service.\n");
        FL_(*to_parse,
                       "Unable to allocate space for new service.");
        if (father_name)
            free(father_name);
        return (FALSE);
    }

    /* set the father to the service */
    if (father_name)
    {
        if (father)
            new_service->father = father;
        new_service->father_name = father_name;
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;

    /* carry on until segment stop or eof, this will handle all lines in current section */
    while ((*to_parse)[0])
    {

        DL_(*to_parse, "parse_service_line, while loop :");

        /* skip spaces and empty lines with them */
        JUMP_NSPACES(*to_parse);
	
	/* end of row */
	if((*to_parse)[0] == ';')
	{
	    (*to_parse)++;
	    continue;
	}

	/* end of file or stack */
	if(!(*to_parse)[0] || (*to_parse)[0] == '}')
	{	
	    DL_(*to_parse, "When escaping from parse_service_line while loop");
	    break;
	}

        /* skip lines starting with '#' */
        if ((*to_parse)[0] == '#')
        {
            REALLY_JUMP_TO_NEXT_ROW(*to_parse);
            continue;
        }

	
        DL_(*to_parse, "parse_service_line, on first option char:");

        /* parse line - search for keywords in g.option_table */
        if (parse_opt(to_parse, type, new_service))
        {
	    if((*to_parse)[0] != ';')
	    {
		FL_(*to_parse, "Must be a ';' char here");
		break;
	    }
	    
	    /* jump the ';' char */
	    (*to_parse)++;
            continue;
        }

        /* Bad content, coud not be parsed */
        /* free and reset */
        initng_service_cache_free(new_service);
        new_service = NULL;
        return (FALSE);
    }

    initng_service_cache_set_string(&FROM_FILE, new_service,
                                    i_strdup(filename));


    if (strcmp(new_service->name, watch_for) == 0)
    {
        D_("Found EXACT match: %s\n", new_service->name);
        (*exact_match) = new_service;
        return (TRUE);
    }

    /* check if this is the service we are looking for. */
    if (service_match(watch_for, new_service->name) == TRUE)
    {
        D_("Found :%s\n", new_service->name);

	/* Just explanation - wi'll erase later (TheLich):
	 * don't keep services with wildcards in service_cache
	 * thus there is each separate copy for each service, that
	 * was created from wilcard service and now all services
	 * in service cache will have unique names (no need in
	 * pattern search when looking for service). This is because
	 * of new behaviour of initng parser (adding to cache only that
	 * service wich was asked). If we will not do that, then if 
	 * wildcard service exists in database and we want to add *real*
	 * service, which matches this wildcard - it will not be added,
	 * wildcard service will be used instead. (daemon/getty/ * and
	 * daemon/getty/1 is a good example).
	*/

	free(new_service->name);
	new_service->name = i_strdup(watch_for);
        (*match) = new_service;
        return (TRUE);
    }

    initng_service_cache_free(new_service);
    
    /* return null if not */
    return (TRUE);
}


#define MAX_OPT_LEN 100
#define MAX_VAR_LEN 100

static int parse_opt(char **where, stype_h * type, service_cache_h * srv)
{
    s_entry *current = NULL;
    char opt_name[MAX_OPT_LEN + 1];
    char *var_name = NULL;
    int opt_len = 0;
    int var_len = 0;

    DL_(*where, "parse_opt");

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "|                               "
     */

    JUMP_NSPACES(*where);
    if (ENDED(*where))
    {
        FL_(*where, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "   |                            "
     */

    /* skip empty services */
    if ((*where)[0] == '}')
    {
        FL_(*where, "found a } char here!");
        return (FALSE);
    }

    /* count number of chars this option name have */
    if ((opt_len = strcspn((*where), "\n; \t=\"'")) < 1)
    {
        FL_(*where, "option name to short!");
        return (FALSE);
    }

    /* get option name and strip all blanks from it */
    if (opt_len >= MAX_OPT_LEN)
    {
        FL_(*where, "Max optlen 100 reached!");
        return (FALSE);
    }
    strncpy(opt_name, (*where), opt_len);
    opt_name[opt_len] = '\0';

    /* walk */
    (*where) += opt_len;
    if (ENDED(*where))
    {
        FL_(*where, "Line just ended after variable name.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "         |                      "
     */

    JUMP_NSPACES(*where);
    if (ENDED(*where))
    {
        FL_(*where, "Line just ended after variable name.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "          |                     "
     */
    DL_(*where, "parse opt, before fetching opt name:");

    if ((*where)[0] != '=' && (*where)[0] != ';')
    {
        /* count number of chars this option name have */
        if (!(var_len = strcspn((*where), "\n; \t=\"'")))
        {
            FL_(*where, "variable name to short!");
            return (FALSE);
        }

        /* get option name and strip all blanks from it */
        if (var_len >= MAX_VAR_LEN)
        {
            FL_(*where, "Maximum variable name of 100 met.");
            return (FALSE);
        }

        /* copy the string */
        var_name = i_strndup(*where, var_len);

        if (!is_valid(var_name))
        {
            FL_(*where,
                           "Variable name contains invalid characters.");
            return (FALSE);
        }

	DL_(*where, var_name);

        (*where) += var_len;
	
	DL_(*where, "EMPTYSPACE");
	
        /* Parsometer
         *
         * "   script start = /usr/bin/gdm; "
         * "               |                "
         */
        JUMP_NSPACES(*where);
        if (ENDED(*where))
        {
            FL_(*where, "Line ended unexpectedly.");
            free(var_name);
            return (FALSE);
        }
        /* Parsometer
         *
         * "   script start = /usr/bin/gdm; "
         * "                |               "
         */
	 

    }

    DL_(*where, " Shud be a '='");

    /* walk the option db */
    while_service_data_types(current)
    {
        /* temporary pointer storage */
        s_entry *tmp = current;

        /* make sure opt_name is set */
        if (!current->opt_name)
            continue;

        /* check that option is not marked for a sertain service type */
        if (current->ot && current->ot != type)
            continue;

        /* check string lenght of word, this is faster then strcasecmp */
        if (opt_len != current->opt_name_len)
        {
            /*F_("Lenght is not correct.\n"); */
            continue;
        }

        /* finaly, check if opt_name is what we are looking for */
        if (strncmp(current->opt_name, opt_name, opt_len) != 0)
            continue;

        D_("parse_opt(%s): option \"%s\" value at option.\n",
           srv->name, current->opt_name);

        /* if this is a alias, browse forward to the correct opt_type */
        while (tmp->opt_type == ALIAS && tmp->alias)
            tmp = tmp->alias;

        D_("option type %i, var_name: %s, opt_name: %s\n", tmp->opt_type,
           var_name ? var_name : "(none)", current->opt_name);

	DL_(*where, current->opt_name);
	
        /* now switch the opt_type */
        switch (tmp->opt_type)
        {
            case STRING:
            case VARIABLE_STRING:
                return(string_parser(current, where, var_name, srv));
            case STRINGS:
            case VARIABLE_STRINGS:
                return(strings_parser(current, where, var_name, srv));
            case SET:
            case VARIABLE_SET:
        	return(set_parser(current, where, var_name, srv));
            case INT:
            case VARIABLE_INT:
                return(iint_parser(current, where, var_name, srv));
            default:
		break;
        }
        FL_(*where,
                               "Unknown, or undefined type to parse!");
	return(FALSE);
    }
    
    /* bail out if we did not find a match */
    FL_(*where, "Did not get a MATCH!");
    F_("Did not get an match on option: \"%s\"!\n", opt_name);

    return (FALSE);
}

static int set_parser(s_entry * type, char **value
                      __attribute__ ((unused)), char *va,
                      service_cache_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    DL_(*value, "set_parser");
    /*DEBUG{
       char print[200];
       sprintf(print, "set parser: type:%s va: %s service: %s", type->opt_name, va, from_service->name);
       FL_(*value, print);
       } */

    /* Parsometer
     *
     * "   enable;"
     * "         |"
     */

    if ((*value)[0] == ';')
    {
        D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
        initng_service_cache_set_var(type, va, from_service);
        return (TRUE);
    }

    /* Parsometer
     *
     * "   enable = yes"
     * "          |    "
     */
    if ((*value)[0] != '=')
    {
        FL_(*value, "There should be an ';' or '=' here!");
        return (FALSE);
    }

    /* skip the '=' char */
    (*value)++;
    JUMP_NSPACES(*value);
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   enable = yes"
     * "            |  "
     */
    if (strncasecmp(*value, "yes", 3) == 0)
    {
        D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
        initng_service_cache_set_var(type, va, from_service);
	
	/* skip 3 chars */
	(*value)+=3;
        return (TRUE);
    }

    if (strncasecmp(*value, "true", 4) == 0)
    {
        D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
        initng_service_cache_set_var(type, va, from_service);

	/* skip 4 chars */
	(*value)+=4;
        return (TRUE);
    }

    /* else, remove */
    initng_service_cache_remove_var(type, va, from_service);
    return (TRUE);
}


static int string_parser(s_entry * type, char **value, char *va,
                         service_cache_h * from_service)
{
    char *to;

    assert(from_service);
    assert(*value);
    assert(value);

    /*FL_(*value, "stringparser"); */
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                |               "
     */

    DL_(*value, "string_parser()");
    if ((*value)[0] != '=')
    {
        FL_(*value, "There should be an = here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                 |              "
     */
    JUMP_NSPACES(*value);
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                  |             "
     */
     
    /* duplicate next word, or string content */
    to=dup_string_and_walk(value, FALSE);

    /* go to the end and stay there */
    while ((*value)[0] && (*value)[0] != ';')
        (*value)++;

    DL_(*value, "skip to the ';' char");

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                              | "
     */
    /* make sure the end is a ';' */
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        free(to);
        return (FALSE);
    }
    
    if ((*value)[0] != ';')
    {
        FL_(*value, "There should be an ; here.");
        if ((*value)[0] == '\n')
            F_("Its a *newline* instead.\n");
        else
            F_("Its a %c instead.\n", (*value)[0]);
        return (FALSE);
    }



    initng_service_cache_set_string_var(type, va, from_service, to);
    return (TRUE);
}

static int iint_parser(s_entry * type, char **value, char *va,
                       service_cache_h * from_service)
{
    char to[11];
    int len = 0;

    assert(from_service);
    assert(*value);
    assert(value);

    /*FL_(*value, "intparser"); */
    /* Parsometer
     *
     * "   pause = 10; "
     * "         |     "
     */

    if ((*value)[0] != '=')
    {
        FL_(*value, "Missing an '=' here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   pause = 10; "
     * "          |    "
     */

    JUMP_NSPACES(*value);
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   pause = 10; "
     * "           |   "
     */

    len = strcspn(*value, "\n;");
    /* make sure the end is a ';' */
    if ((*value)[len] != ';')
    {
        FL_(*value, "Missing an ';' char.");
        return (FALSE);
    }

    /* make sure value is not to big */
    if (len >= 10)
    {
        FL_(*value, "Length is bigger than 10.");
        return (FALSE);
    }

    /* ok, copy the code */
    strncpy(to, *value, len);
    to[len] = '\0';

    /* walk */
    (*value) += len;
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    initng_service_cache_set_int_var(type, va, from_service, atoi(to));
    return (TRUE);
}

static int strings_parser(s_entry * type, char **value, char *va,
                          service_cache_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    assert(*value);
    assert(value);

    /*FL_(*value, "stringS parser"); */
    /* Parsometer
     *
     * "   need = test service; "
     * "        |               "
     */
    DL_(*value, "strings_parser()");


    if ((*value)[0] != '=')
    {
        FL_(*value, "Missing an '=' here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   need = test service; "
     * "         |              "
     */
    JUMP_NSPACES(*value);
    if (ENDED(*value))
    {
        FL_(*value, "Line ended unexpectedly.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   need = test service; "
     * "          |             "
     */
    DL_(*value, "strings_parser, Youst before dup_string_and_walk(value, TRUE)");

    while ((to = dup_string_and_walk(value, TRUE)))
    {
	DL_(*value, "adding");
	if(va)
	{
    	    initng_service_cache_set_another_string_var(type, i_strdup(va), from_service,
                                                   to);
	} else {
    	    initng_service_cache_set_another_string_var(type, NULL, from_service,
                                                    to);
	}
	/* check if this is end */
	if ((*value)[0]==';')
	    break;

        JUMP_NSPACES(*value);

	/* check if this is end */
	if ((*value)[0]==';')
	    break;
	if (ENDED(*value))
	    break;
	DL_(*value, "a_add");
    }
    
    /* dont forgot to free va */
    if(va)
    {
	free(va);
	va=NULL;
    }
    /* Parsometer
     *
     * "   need = test service; "
     * "                      | "
     */

    DL_(*value, "all added");
    if ((*value)[0] != ';')
    {
        FL_(*value, "End char is not an ';'.");
        return (FALSE);
    }
    return (TRUE);
}

int module_init(const char *version)
{

    D_("i_parser: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    return (initng_plugin_hook_add(PARSERS, 50, &initng_i_parser));

}

void module_unload(void)
{
    D_("i_parser: module_unload();\n");
    initng_plugin_hook_del(PARSERS, &initng_i_parser);
}


static char *dup_string_and_walk(char **value, int space_is_new_word)
{
    char *to = NULL;
    int len = 0;
    /* Parsometer
     *
     * "   script start = { /usr/bin/gdm };            "
     * "   script start = "/usr/bin/gdm";              "
     * "   script start = /usr/bin/gdm;                "
     * "   script start = /usr/bin/gdm /usr/sbin/gdm;  "
     * "                  |                            "
     */


	    if(space_is_new_word==TRUE)
	    {	
		DL_(*value, "Space is new word");
	    } else {
		DL_(*value, "Space is none word");	    
	    }

    DL_(*value, "dup_string_and_walk:");
    
    
    
    /* If this value we shud copy is stacked in braces */
    if ((*value)[0] == '{')
    {
        int stack = 1;

        (*value)++;

#ifndef NOTRIM
        JUMP_NSPACES(*value);
#endif
        /* count lenght to the '}' char */
        while ((*value)[len])
        {
            if ((*value)[len] == '{')
                stack++;
            else if ((*value)[len] == '}')
                stack--;

	    /* if stack is zero or below, we are standing on a '}' with no excape char before */
            if (stack <= 0 && (*value)[len] == '}' && (*value)[len-1] != '\\')
            {
                (*value)[len] = ';';
                break;
            }
            len++;
        }

#ifndef NOTRIM
	while((*value)[len-1]==' ' || (*value)[len-1]=='\t' || (*value)[len-1]=='\n')
	    len--;
#endif
	/* ok, copy the code */
	if (!(to = i_strndup(*value, len)))
	{
    	    FL_(*value, "Failed copying.");
    	    return (NULL);
	}

	/* walk forward again */
	(*value) += len;
	
#ifndef NOTRIM	
	JUMP_NSPACES(*value);
#endif
	if((*value)[0]=='}')
	    (*value)++;
	    
	DL_(*value, "after += len");
	
	return(to);

    }
    
    
    /* if it is embraced in with '"' */
    if ((*value)[0] == '"')
    {
	/* skip the '"' char */
	(*value)++;
	
	/* count the number of tokens */
	while ((*value)[len])
	{
	    /* if this is the ending '"' with no escapechar before */
	    if ((*value)[len] == '"' && (*value)[len-1] != '\\')
	    {
		break;
	    }
	    len++;
	}

	/* ok, copy the code */
	if (!(to = i_strndup(*value, len)))
	{
    	    FL_(*value, "Failed copying.");
    	    return (NULL);
	}

	/* walk forward again */
	(*value) += len;
	if((*value)[0]=='"')
	    (*value)++;
	
	DL_(*value, "after += len");
	
	return(to);
    }

    /* else */    
    DL_(*value, "Non stack or quoted value");

    /* Count lenght */
    while((*value)[len])
    {
	    /* this char is escaped by the char before */
	    if((*value)[len-1] == '\\')
	    {
		len++;
		continue;
	    }
	    
	    /* If we shud break copying on a space, check this */
	    if(space_is_new_word==TRUE)
	    {	
		DL_(*value, "Space is new word");
		if((*value)[len]==' ' || (*value)[len]=='\t' || (*value)[len]=='\n')
		    break;
	    }
	
	    /* always break on a ; */
	    if((*value)[len]==';')
		break;
		
	    len++;
    }

    /* ok, copy the code */
    if (!(to = i_strndup(*value, len)))
    {
        FL_(*value, "Failed copying.");
        return (FALSE);
    }

    /* walk forward again */
    (*value) += len;
    DL_(*value, "after += len");
    
    return(to);
}
