#include <stdlib.h>
#include "ngcs_common.h"
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

int ngcs_writemsg(int sock, int chan, ngcs_data * data, int cnt)
{
    int head[2];
    int n;

    head[0] = chan;
    head[1] = cnt;
    if (ngcs_sendall(sock, head, 2 * sizeof(int)))
        return 1;
    for (n = 0; n < cnt; n++)
    {
        head[0] = data[n].type;
        switch (data[n].type)
        {
            case NGCS_TYPE_GROUP:
            case NGCS_TYPE_INT:
                head[1] = sizeof(int);
                if (ngcs_sendall(sock, head, 2 * sizeof(int)))
                    return 1;
                if (ngcs_sendall(sock, &data[n].d.i, sizeof(int)))
                    return 1;
                break;
            case NGCS_TYPE_LONG:
                head[1] = sizeof(long);
                if (ngcs_sendall(sock, head, 2 * sizeof(int)))
                    return 1;
                if (ngcs_sendall(sock, &data[n].d.l, sizeof(long)))
                    return 1;
                break;
            case NGCS_TYPE_STRING:
                if (data[n].size < 0)
                    data[n].size = strlen(data[n].d.s);
                head[1] = data[n].size;
                if (ngcs_sendall(sock, head, 2 * sizeof(int)))
                    return 1;
                if (ngcs_sendall(sock, data[n].d.s, data[n].size))
                    return 1;
                break;
            default:
                if (data[n].size < 0)
                    return 1;
                head[1] = data[n].size;
                if (ngcs_sendall(sock, head, 2 * sizeof(int)))
                    return 1;
                if (ngcs_sendall(sock, data[n].d.p, data[n].size))
                    return 1;
                break;
        }
    }
    return 0;
}

int ngcs_readmsg(int sock, int *chan, ngcs_data ** data, int *cnt)
{
    int head[2];
    int n;
    if (ngcs_recvall(sock, head, 2 * sizeof(int)))
        return 1;
    *chan = head[0];
    *cnt = head[1];
    *data = NULL;
    if (*cnt <= 0)
        return 0;
    *data = malloc((*cnt) * sizeof(ngcs_data));
    if (!*data)
        return 1;
    for (n = 0; n < *cnt; n++)
    {
        if (ngcs_recvall(sock, head, 2 * sizeof(int)))
            return 1;
        if (head[1] < 0)
            return 1;
        (*data)[n].type = head[0];
        (*data)[n].size = head[1];
        switch ((*data)[n].type)
        {
            case NGCS_TYPE_GROUP:
            case NGCS_TYPE_INT:
                if ((*data)[n].size != sizeof(int))
                    return 1;
                if (ngcs_recvall(sock, &(*data)[n].d.i, sizeof(int)))
                    return 1;
                break;
            case NGCS_TYPE_STRING:
                (*data)[n].d.s = malloc((*data)[n].size + 1);
                if (!(*data)[n].d.s)
                    return 1;
                if (ngcs_recvall(sock, (*data)[n].d.s, (*data)[n].size))
                    return 1;
                (*data)[n].d.s[(*data)[n].size] = 0;
                break;
            default:
                (*data)[n].d.p = malloc((*data)[n].size);
                if (!(*data)[n].d.p)
                    return 1;
                if (ngcs_recvall(sock, (*data)[n].d.p, (*data)[n].size))
                    return 1;
                break;
        }
    }
    return 0;
}

void ngcs_freedata(ngcs_data * data, int cnt)
{
    int i;

    if (!data || cnt < 0)
        return;
    for (i = 0; i < cnt; i++)
    {
        switch (data[i].type)
        {
            case NGCS_TYPE_INT:
                break;
            case NGCS_TYPE_STRING:
                if (data[i].d.s)
                    free(data[i].d.s);
                break;
            default:
                if (data[i].d.p)
                    free(data[i].d.p);
                break;
        }
    }
    free(data);
}

int ngcs_sendall(int sock, const void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = send(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}

int ngcs_recvall(int sock, void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = recv(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}
