
/* Initng, a 6next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#define _GNU_SOURCE
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>

#include "../../src/initng_toolbox.h"
#include "../../src/initng_global.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_active_state.h"

#include "print_service.h"

#define IS_PRINTABLE(x) (x >= 32 || x == '\n' || x == '\t' || x == '\r')

static void service_db_print_u(service_cache_h * s, FILE * fd, int intro);
static void active_db_print_u(active_db_h * s, FILE * fd, int intro);
static void print_string_value(char *string, FILE * fd);

/* P R I N T   S E R V I C E */

/* some stdout */
void service_db_print(service_cache_h * s, FILE * fd)
{
    service_db_print_u(s, fd, TRUE);
    return;
}

void active_db_print(active_db_h * s, FILE * fd)
{
    active_db_print_u(s, fd, TRUE);
    return;
}

/* there may be unprintable characters in string - should escape them when printing */
static void print_string_value(char *string, FILE * fd)
{
    int i;

    for (i = 0; string[i] != 0; i++)
    {
        if (IS_PRINTABLE(string[i]))
            fprintf(fd, "%c", string[i]);
        else
            fprintf(fd, "^%c", string[i] ^ 0x40);
    }
}

static void service_db_print_u(service_cache_h * s, FILE * fd, int intro)
{
    /*data path */
    s_data *tmp = NULL;

    assert(s);
    assert(s->name);

    /* try calculate out how mutch memory we wanna allocate
       that we will send */

    if (intro)
    {
        fprintf(fd, "\n#  %s  \"%s", s->type->name, s->name);
        if (s->father_name)
            fprintf(fd, " : %s", s->father_name);
        fprintf(fd, "\"\n");
    }

    fprintf(fd, "    ------------------------ \n");

    list_for_each_entry(tmp, &s->data.list, list)
    {
        if (!tmp->type)
            continue;
        switch (tmp->type->opt_type)
        {
            case STRING:
            case STRINGS:
                if (!tmp->t.s)
                {
                    F_("empty value!\n");
                    continue;
                }
                fprintf(fd, " @@ %s:\t\"", tmp->type->opt_name);
                print_string_value(tmp->t.s, fd);
                fprintf(fd, "\"\n");
                continue;
            case VARIABLE_STRING:
            case VARIABLE_STRINGS:
                if (!tmp->t.s)
                {
                    F_("empty value!\n");
                    continue;
                }
                fprintf(fd, " @@ %s \"%s\" : \"", tmp->type->opt_name, tmp->vn);
                print_string_value(tmp->t.s, fd);
                fprintf(fd, "\"\n");
                continue;
            case INT:
                fprintf(fd, " @@ %s:\t\"%i\"\n", tmp->type->opt_name,
                        tmp->t.i);
                continue;
            case VARIABLE_INT:
                fprintf(fd, " @@ %s %s:\t\"%i\"\n", tmp->type->opt_name,
                        tmp->vn, tmp->t.i);
                continue;
            case SET:
                fprintf(fd, " @@ %s\n", tmp->type->opt_name);
                continue;
            case VARIABLE_SET:
                fprintf(fd, " @@ %s %s\n", tmp->type->opt_name, tmp->vn);
                continue;
            case ALIAS:
                fprintf(fd, " @@ ALIAS %s\n", tmp->type->opt_name);
                continue;
            case U_D_T:
                continue;
        }
    }

    /* if father with data exits, print it */
    if (s->father)
    {
        D_("FATHER FOUND %s\n", s->name);
        service_db_print_u(s->father, fd, FALSE);
    }
    else
        D_("FATHER NOT FOUND %s\n", s->name);

    return;
}

static void print_max_20(char *from, FILE * fd)
{
    int i = 0;
    char to_pr[50];

    assert(from);

    while (from[0] == ' ' || from[0] == '\n' || from[0] == '\t')
        from++;
    while (i <= 40 && from[i] && from[i] != '\n')
        i++;
    strncpy(to_pr, from, i);
    to_pr[i] = '\0';
    if (i >= 40)
        fprintf(fd, "%s ...", to_pr);
    else
        fprintf(fd, "%s", to_pr);
}

/* Walk through every service and print it all */
void service_db_print_all(FILE * fd)
{

    service_cache_h *current = NULL;

    fprintf(fd, "Full service print:\n");

    D_("print_all():\n");

    while_service_cache(current)
    {
        D_("printing %s ...\n", current->name);
        service_db_print(current, fd);
    }

    return;
}
static void active_db_print_process(process_h * p, FILE * fd)
{
    assert(p);
    fprintf(fd, "\tProcess: type %s\n", p->pt->name);
    fprintf(fd, "\t\tPid: %i\n", p->pid);
    fprintf(fd, "\t\tR_code: %i\n", p->r_code);
    fprintf(fd, "\t\tFds read: %i, write: %i\n", p->out_pipe[0],
            p->out_pipe[1]);
    if (p->buffer)
    {
        fprintf(fd, "\t\tBuffer: \"");
        print_max_20(p->buffer, fd);
        fprintf(fd, "\"\n");
    }
    fprintf(fd, "\t\tBuffer-pos: %i\n\n", p->buffer_pos);

}

static void active_db_print_u(active_db_h * s, FILE * fd, int intro)
{
    /*data path */
    s_data *tmp = NULL;
    process_h *process = NULL;

    assert(s);
    assert(s->name);

    /* try calculate out how mutch memory we wanna allocate
       that we will send */

    if (intro)
    {
        struct timeval now;

        fprintf(fd, "\n#  %s  \"%s", s->type->name, s->name);
        if (s->from_service && s->from_service->father_name)
            fprintf(fd, " : %s", s->from_service->father_name);
        fprintf(fd, "\"\n");


        if (s->current_state && s->current_state->state_name)
        {
            fprintf(fd, "\tactive status: %s\n",
                    s->current_state->state_name);
        }

        gettimeofday(&now, NULL);

        fprintf(fd,
                "\tlast_rought_time: %ims time_last_state: %ims time_current_state: %ims\n",
                MS_DIFF(now, s->last_rought_time), MS_DIFF(now,
                                                           s->
                                                           time_last_state),
                MS_DIFF(now, s->time_current_state));

        /* print processes if any */
        while_processes(process, s)
        {
            active_db_print_process(process, fd);
        }

    }




    fprintf(fd, "\tactive_db opts:\n");

    list_for_each_entry(tmp, &(s->data.list), list)
    {
        if (tmp->type)
        {
            switch (tmp->type->opt_type)
            {
                case STRING:
                case STRINGS:

                    if (!tmp->t.s)
                        break;
                    fprintf(fd, "\t\t@@ %s:\t\"", tmp->type->opt_name);
                    print_max_20(tmp->t.s, fd);
                    fprintf(fd, "\"\n");
                    break;

                case INT:
                    fprintf(fd, "\t\t@@ %s:\t\"%i\"\n", tmp->type->opt_name,
                            tmp->t.i);
                    break;

                case SET:
                    fprintf(fd, "\t\t@@ %s\n", tmp->type->opt_name);
                    break;

                default:
                    break;

            }
        }
    }

    if (s->from_service)
    {
        fprintf(fd, "\tservice_db opts:\n");
        tmp = NULL;
        list_for_each_entry(tmp, &(s->from_service->data.list), list)
        {
            if (tmp->type)
            {
                switch (tmp->type->opt_type)
                {
                    case STRING:
                    case STRINGS:
                    case VARIABLE_STRING:
                    case VARIABLE_STRINGS:

                        if (!tmp->t.s)
                            break;
                        fprintf(fd, "\t\t@@ %s:\t\"%s\"\n", tmp->type->opt_name, tmp->t.s);
                        break;

                    case INT:
                    case VARIABLE_INT:
                        fprintf(fd, "\t\t@@ %s:\t\"%i\"\n",
                                tmp->type->opt_name, tmp->t.i);
                        break;
                    case SET:
                    case VARIABLE_SET:
                        fprintf(fd, "\t\t@@ %s\n", tmp->type->opt_name);
                        break;
                    default:
                        break;
                }
            }
        }
    }
}

/* Walk through every service and print it all */
void active_db_print_all(FILE * fd)
{

    active_db_h *apt = NULL;

    fprintf(fd, "Full service print:\n");

    D_("print_all():\n");

    while_active_db(apt)
    {
        active_db_print(apt, fd);
        fprintf(fd, "\n\n");
    }

    return;
}
