
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <CUnit/CUnit.h>
#include <CUnit/Basic.h>
#include "test_initng.h"

#include <stdlib.h>
#include <string.h>
#include "../src/initng_string_tools.h"

static void test_split_n(void)
{
    char *to_split = NULL;
    char **splitted = NULL;
    int len;

    NEW_TEST();

    TEST("Try to split a long string set.");
    to_split = strdup("First\nSecond,Third;Forth Fifth");
    CU_ASSERT(to_split != NULL);
    if (to_split == NULL)
        return;

    splitted = split_n(to_split, &len, 0, 0);
    CU_ASSERT(splitted != NULL);
    if (splitted == NULL)
        return;

    CU_ASSERT_STRING_EQUAL(splitted[0], "First");
    CU_ASSERT_STRING_EQUAL(splitted[1], "Second");
    CU_ASSERT_STRING_EQUAL(splitted[2], "Third");
    CU_ASSERT_STRING_EQUAL(splitted[3], "Forth");
    CU_ASSERT_STRING_EQUAL(splitted[4], "Fifth");


    /* free data */
    free(splitted);
    free(to_split);
}

static void test_chars_to_end_of_this_stack(void)
{
    char *string = NULL;

    NEW_TEST();

    TEST("test stack count.");
    string = strdup("this is a stack\n} out of it");
    CU_ASSERT(chars_to_end_of_this_stack(string) == 16);
    free(string);

    TEST("test stack count, with an inside stack.");
    string = strdup("this { inside } stack} out");
    CU_ASSERT(chars_to_end_of_this_stack(string) == 21);
    free(string);

}

static void test_st_cmp(void)
{
    char *string = NULL;
    char *a = NULL;

    NEW_TEST();

    TEST("st_cmp.");
    string = strdup("good test2 test3 \"test 4\"");
    a = string;
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;


    TEST("cmp good");
    CU_ASSERT(st_cmp(&a, "good") == TRUE);
    TEST("cmp baad");
    CU_ASSERT(st_cmp(&a, "baad") == FALSE);
    TEST("cmp test2");
    CU_ASSERT(st_cmp(&a, "test2") == TRUE);
    TEST("cmp bad_hit");
    CU_ASSERT(st_cmp(&a, "bad_hit") == FALSE);
    TEST("cmp test3");
    CU_ASSERT(st_cmp(&a, "test3") == TRUE);
    TEST("cmp \"test 4\"");
    CU_ASSERT(st_cmp(&a, "test 4") == TRUE);

    free(string);

}

static void test_st_dup_next_word(void)
{
    char *string = NULL;
    char *a = NULL;
    char *copy = NULL;

    NEW_TEST();

    TEST("st_cmp()");
    string = strdup("good test2 test3 \"test 4\"");
    a = string;
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;

    TEST("copy string");
    copy = st_dup_next_word(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "good");
        free(copy);
        copy = NULL;
    }

    TEST("copy next string");
    copy = st_dup_next_word(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "test2");
        free(copy);
        copy = NULL;
    }

    TEST("copy third string");
    copy = st_dup_next_word(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "test3");
        free(copy);
        copy = NULL;
    }

    TEST("copy string with \"sp ac es\"");
    copy = st_dup_next_word(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "test 4");
        free(copy);
        copy = NULL;
    }

}

static void test_JUMP_ROW_and_SPACES(void)
{
    char *string = NULL;
    char *a = NULL;

    NEW_TEST();

    TEST("set a long string ..");
    string = strdup("This is row 1    This is row 2; This is row 3;\t \tThis is row 4\n This is row 5\n");
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;
    a = string;

    TEST("check row 1.");
    CU_ASSERT(strncmp("This is row 1", a, strlen("This is row 1")) == 0);
    a += strlen("This is row 1");

    TEST("check row 2.");
    JUMP_SPACES(a);
    CU_ASSERT(strncmp("This is row 2", a, strlen("This is row 2")) == 0);

    TEST("check row 3.");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    CU_ASSERT(strncmp("This is row 3", a, strlen("This is row 3")) == 0);

    TEST("check row 4.");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    CU_ASSERT(strncmp("This is row 4", a, strlen("This is row 4")) == 0);

    TEST("check row 5.");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    CU_ASSERT(strncmp("This is row 5", a, strlen("This is row 5")) == 0);

    free(string);
}

static void test_st_dup_line(void)
{
    char *string = NULL;
    char *a = NULL;
    char *copy = NULL;

    NEW_TEST();

    TEST("set a long string ..");
    string = strdup("This is row 1\n  This is row 2; This is row 3;\t \tThis is row 4\n This is row 5\n");
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;
    a = string;

    TEST("check row 1");

    copy = st_dup_line(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "This is row 1");
        free(copy);
        copy = NULL;
    }


    TEST("check row 2");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    copy = st_dup_line(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "This is row 2");
        free(copy);
        copy = NULL;
    }

    TEST("check row 3");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    copy = st_dup_line(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "This is row 3");
        free(copy);
        copy = NULL;
    }

    TEST("check row 4");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    copy = st_dup_line(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "This is row 4");
        free(copy);
        copy = NULL;
    }

    TEST("check row 5");
    JUMP_TO_NEXT_ROW(a);
    JUMP_SPACES(a);
    copy = st_dup_line(&a);
    CU_ASSERT(copy != NULL);
    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "This is row 5");
        free(copy);
        copy = NULL;
    }



    free(string);
}


void test_st_strip_path(void)
{
    char *string = NULL;
    char *copy = NULL;

    NEW_TEST();

    TEST("set \"test/to/go\" check that we got \"go\"");
    string = strdup("test/to/go");
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;

    copy = st_strip_path(string);
    CU_ASSERT(copy != NULL);

    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "go");
        free(copy);
    }

    free(string);
}

void test_st_strip_end(void)
{
    char *string = NULL;

    NEW_TEST();

    TEST("set \"test/to/go\" check that we got \"test/to\"");
    string = strdup("test/to/go");
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;

    st_strip_end(&string);
    CU_ASSERT(string != NULL);
    if (!string)
        return;

    CU_ASSERT_STRING_EQUAL(string, "test/to");
    TEST("strip again to \"test\"");
    st_strip_end(&string);
    CU_ASSERT(string != NULL);
    if (!string)
        return;
    CU_ASSERT_STRING_EQUAL(string, "test");
    free(string);
}

void test_st_get_path(void)
{
    char *string = NULL;
    char *copy = NULL;

    NEW_TEST();

    TEST("set \"test/to/go\" check that we got \"test/to\"");
    string = strdup("test/to/go");
    CU_ASSERT(string != NULL);
    if (string == NULL)
        return;

    copy = st_get_path(string);
    CU_ASSERT(copy != NULL);

    if (copy)
    {
        CU_ASSERT_STRING_EQUAL(copy, "test/to");
        free(copy);
    }

    free(string);
}


void test_initng_string_tools_setup(void)
{
    CU_pSuite suite;

    suite = CU_add_suite("initng_string_tools", empty_init, empty_cleanup);
    CU_add_test(suite, "split_n", test_split_n);
    CU_add_test(suite, "chars_to_end_of_this_stack",
                test_chars_to_end_of_this_stack);
    CU_add_test(suite, "st_cmp", test_st_cmp);
    CU_add_test(suite, "dup_next_word", test_st_dup_next_word);
    CU_add_test(suite, "JUMP_ROW_and_SPACES", test_JUMP_ROW_and_SPACES);
    CU_add_test(suite, "st_dup_line", test_st_dup_line);
    CU_add_test(suite, "st_strip_path", test_st_strip_path);
    CU_add_test(suite, "st_strip_end", test_st_strip_end);
    CU_add_test(suite, "st_get_path", test_st_get_path);
}
